import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spl
import scipy.linalg
sqrtm = scipy.linalg.sqrtm
norm = np.linalg.norm
inv = np.linalg.inv
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

eta = 0.001
mup = 0.01
mud = 0.1
Nx = 20
Ny = Nx-1

A = sp.csc_matrix(sp.diags([np.array([1] +  [1+eta] * (Nx-1)), -np.ones(Nx)], [0, -1])).toarray()
A = A[1:]
Q = mup * np.eye(Nx)
S = mud * np.eye(Ny)

b = np.array([1] + [0]*(Ny-1))
c = np.zeros(Nx)

gamma = 0.95
tau = 0.1
sigma = gamma/tau/norm(A)**2


M1 = np.vstack([np.hstack([Q, A.T]),
                np.hstack([A, -S])])
HessL = M1

M = np.vstack([np.hstack([Q / tau, A.T / np.sqrt(sigma*tau)]),
               np.hstack([A / np.sqrt(sigma*tau), S / sigma])])
u, s, v = spl.svds(M1, 1)
NM = s[0] ** 2

msr = min(np.abs(spl.eigs(M1, which = 'SM')[0]))
print('eta =', eta, 'mup =', mup, 'mud =', mud,  'msr =', msr)


def L(x, y):
    return 0.5 * x @ Q @ x + y @ (A @ x - b) - 0.5 * y @ S @ y


Ix = np.eye(Nx)
Iy = np.eye(Ny)

def recursion_matrix(tau, sigma):
    assert np.abs(sigma - gamma/tau/norm(A)**2) < 1e-10, [tau, sigma, gamma/tau/norm(A)**2]
    R = np.vstack([np.hstack([inv(Ix+tau*Q) @ (Ix - 2*tau*sigma * A.T @ inv(Iy+sigma*S) @ A),
                          tau * inv(Ix+tau*Q) @ A.T @ (Iy - 2 * inv(Iy+sigma*S))]),
               np.hstack([sigma * inv(Iy+sigma*S) @ A,
                          inv(Iy+sigma*S)])])
    
    Rb = np.hstack([-tau*inv(Ix+tau*Q) @ (-2 * sigma * A.T @ inv(Iy+sigma*S) @ b + c),
                -sigma*inv(Iy+sigma*S) @ b])
    return R, Rb

R, Rb = recursion_matrix(tau, sigma)

def visualize_eigenvalues(TAUS):
    for tau in TAUS:
        sigma = gamma / tau / norm(A)**2
        R, Rb = recursion_matrix(tau, sigma)
        w, vl, vr = scipy.linalg.eig(R, left=True, right=True)

        plt.plot(np.real(w), np.imag(w), '.')
    plt.legend(TAUS)
    plt.show()
    return

def deriv_recursion_matrix(tau):
    dsigma_dtau = -sigma/tau
    dR_dtau = np.vstack([np.hstack([-inv(Ix+tau*Q) @ Q @ inv(Ix+tau*Q) @ (Ix - 2*tau*sigma * A.T @ inv(Iy+sigma*S) @ A)
                                + inv(Ix+tau*Q) @ (2*tau*sigma * A.T @ inv(Iy+sigma*S)@S@inv(Iy+sigma*S) @ A) * dsigma_dtau,
                          1/tau**2 * inv(Ix/tau+Q) @ inv(Ix/tau+Q) @ A.T @ (Iy - 2 * inv(Iy+sigma*S))
                                    + 2 * inv(Ix/tau+Q) @ A.T @ inv(Iy+sigma*S)@S@inv(Iy+sigma*S)*dsigma_dtau]),
               np.hstack([-inv(Iy/sigma+S) @ inv(Iy/sigma+S) @ A * (-dsigma_dtau)/sigma**2,
                          -inv(Iy+sigma*S) @ S @ inv(Iy+sigma*S) * dsigma_dtau])])
    return dR_dtau


def deriv_eigenvalue(tau, sigma=None):
    # positive implies decrease tau, negative implies increase tau
    if sigma == None:
        sigma = gamma / tau / norm(A)**2
    R, Rb = recursion_matrix(tau, sigma)
    w, vl, vr = scipy.linalg.eig(R, left=True, right=True)
    i = np.argmax(np.abs(w) * (np.abs(w) < 1))  # we remove the eigenvalue 1
    w = w[i]

    # wrong formula
    #vl = vl[:,i]
    #vr = vr[:,i]
    #dR_dtau = deriv_recursion_matrix(tau)
    #dw_dtau = (vl.conj() @ dR_dtau @ vr) / (vl.conj() @ vr)
    
    tau_eps = tau * (1+1e-7)
    Reps, Rbeps = recursion_matrix(tau_eps, gamma / tau_eps / norm(A)**2)
    w_eps, vl, vr = scipy.linalg.eig(Reps, left=True, right=True)
    w_eps = w_eps[i]
    dw_dtau = (w_eps - w)/(tau_eps - tau)
    return w, vl, vr, dw_dtau

def primal_dual_residuals_ratio(tau, z0, K=100):
    # larger than 1.5 implies increase tau, smaller than 1/1.5 implies decrease tau
    sigma = gamma/tau/norm(A)**2
    x, y, x_in, y_in, Nprimal, Ndual, tV = pdhg(z0[:Nx], z0[Nx:], tau, sigma, K)
    z = np.concatenate([x,y])

    res = M1 @ z - np.hstack([-c, b])

    return res, norm(res), norm(res[:Nx], 1) / norm(res[Nx:], 1)

def deriv_squared_residual(tau, z0, z1=None):
    # derivative of ||nabla L(T(z))||**2 with respect to tau
    if z1 == None:
        x, y, x_in, y_in, Nprimal, Ndual, tV = pdhg(z0[:Nx], z0[Nx:], tau, sigma, 1)
        z1 = np.concatenate([x,y])
    res = M1 @ z1 - np.hstack([-c, b])
    return 2 * res @ HessL @ deriv_recursion_matrix(tau) @ z0


def adaptive_steps_goldstein(x, y, x_, y_, tau, sigma, memory, eta=0.95):
    Delta = 1.5
    if len(memory) == 0:
        alpha = 0.5
    else:
        alpha = memory[-1][2]

    primal_residual = np.linalg.norm((x - x_)/tau
                                     - A.T @ (y - y_), 1)
    dual_residual = np.linalg.norm((y - y_)/sigma
                                   - A @ (x - x_), 1)

    up_down = 0
    if primal_residual > Delta * dual_residual:
        up_down = 1
        alpha = alpha * eta
    elif dual_residual > Delta * primal_residual:
        up_down = -1
        alpha = alpha * eta
    
    sigma = sigma * (1-alpha)**up_down
    tau = tau / (1-alpha)**up_down
    memory.append([tau, sigma, alpha, primal_residual, dual_residual])
    return tau, sigma


def adaptive_steps_derivative(x, y, x_, y_, tau, sigma, memory):
    w, vl, vr, dw_dtau = deriv_eigenvalue(tau, sigma)
    if 1:
        if np.real(dw_dtau * np.conj(w)) > 0:
            up_down = 1
        else:
            up_down = -1
        if len(memory) == 0:
            alpha = 0.5
        elif (len(memory) > 0 and up_down == memory[-1][3]):
            alpha = min(0.5, memory[-1][2])
        else:  # up_down is changing
            alpha = max(0.0001, memory[-1][2] * 0.5)

        tau = tau * ((1-alpha) ** up_down)
        sigma = sigma / ((1-alpha) ** up_down)

    memory.append([tau, sigma, alpha, up_down])
    return tau, sigma


def determine_largest_eigenvalues(R, z, n):
    Z = np.zeros((len(z), n+1))
    Z[:, 0] = z
    for i in range(1, n+1):
        Z[:, i] = R @ Z[:, i-1]
    p, _, _, _ = np.linalg.lstsq(Z[:, :n], -Z[:, n], rcond=None)
    p = np.polynomial.Polynomial(np.concatenate([p, [1]]))
    eigs = p.roots()
    return eigs

def determine_2_largest_eigenvalues(R, x, y, tau, sigma):
    Nx = len(x)
    z0 = np.concatenate([x,y])
    z1 = R @ z0
    z2 = R @ z1
    res00 = norm(x)**2/tau + norm(y)**2/sigma
    res01 = z0[:Nx] @ z1[:Nx] / tau + z0[Nx:] @ z1[Nx:] / sigma
    res11 = norm(z1[:Nx])**2 / tau + norm(z1[Nx:])**2 / sigma
    res02 = z0[:Nx] @ z2[:Nx] / tau + z0[Nx:] @ z2[Nx:] / sigma
    res22 = norm(z2[:Nx])**2 / tau + norm(z2[Nx:])**2 / sigma
    res12 = z1[:Nx] @ z2[:Nx] / tau + z1[Nx:] @ z2[Nx:] / sigma

    # we start with many equations: p2 z2 + p1 z1 = z0 + O(lambda3**s)
    # we simplify into two equations:
    #               ||z0||_V**2 = ||p1 z1 + p2 z2||_V**2
    #               ||z0||_V**2 = z0.T DV (p1 z1 + p2 z2)
    # this becomes
    # p2 = (res00 - p1 * res01) / res02
    # res00 = p1**2 * res11 + p2**2 * res22 + 2*p1*p2*res12
    # 0 = - res00 + p1**2 * res11 + ((res00 - p1 * res01) / res02)**2 * res22
    #             + 2*p1*(res00 - p1 * res01) / res02*res12
    # 0 = - res00 + p1**2 * res11 + (res00**2 + p1**2 * res01**2 - 2 * p1*res01*res00) * res22/res02**2 + 2*(p1*res00 - p1**2 * res01)*res12/res02
    p = np.polynomial.Polynomial([-res00+res00**2*res22/res02**2,
                                  -2*res01*res00*res22/res02**2+2*res00*res12/res02,
                                  res11+res01**2*res22/res02**2-2*res01*res12/res02])
    #p1 = p.roots()
    #p2 = (res00 - p1 * res01) / res02
    p1 = (res00*res12/res02 - res01) / (res01*res12/res02 - res11)
    p2 = (res00 - p1 * res01) / res02
    eigs = np.polynomial.Polynomial([1, p1, p2]).roots()
    return eigs


def approximate_spectral_radius(res):
    rres = np.real(res[1:]/res[:-1])
    sign_diff_rres = np.sign(np.diff(rres))
    current = sign_diff_rres[-1]
    first = 0
    second = 0
    for i in range(-2,-len(res)+1,-1):
        if sign_diff_rres[i] != current:
            if first == 0:
                first = i
                # make it more precise by fitting a quadratic model
                p = rres[i-1] / (-1*-2) * np.polynomial.Polynomial.fromroots([0, 1]) \
                    + rres[i] / (1*-1) * np.polynomial.Polynomial.fromroots([-1, 1]) \
                    + rres[i+1] / (1*2) * np.polynomial.Polynomial.fromroots([-1, 0])
                first = i-p.coef[1] / 2. / p.coef[2]
                current = -current
            else:
                second = i
                # make it more precise by fitting a quadratic model
                p = rres[i-1] / (-1*-2) * np.polynomial.Polynomial.fromroots([0, 1]) \
                    + rres[i] / (1*-1) * np.polynomial.Polynomial.fromroots([-1, 1]) \
                    + rres[i+1] / (1*2) * np.polynomial.Polynomial.fromroots([-1, 0])
                second = i-p.coef[1] / 2. / p.coef[2]
                break
    if second == 0:
        return np.sqrt(rres[-1])
    else:
        # midpoint = (first + second + 1)//2
        # return np.sqrt(rres[midpoint])
        midpoint = (first + second) / 2.
        midpoint0 = int(np.floor(midpoint))
        return np.sqrt(rres[midpoint0] * (midpoint-midpoint0) + rres[midpoint0+1] * (midpoint0+1 - midpoint))


def adaptive_steps_residual_norm(x, y, x_, y_, tau, sigma, memory,
                                 goldstein_warm_up=False):
    # We mimick the method on derivatives of the eigenvalues but with the power method instead.

    if len(memory) > 0:
        up_down = memory[-1][12]
    else:
        up_down = 1
        
    alpha = 2./3.
    delta = 0.6
    # 1. We estimate the rate of convergence using residual norm
    res_x = norm(x_ - x) ** 2
    res_y = norm(y_ - y) ** 2
    if len(memory) > 0 and memory[-1][0] == tau:
        counter = memory[-1][5] + 1
    else:
        counter = 0

    estimated_rate_1 = np.nan  # default value
    estimated_rate_2 = np.nan  # default value
    estimated_rate_3 = np.nan  # default value
    estimated_rate_4 = np.nan
    estimated_rate_5 = np.nan
    estimated_rate = np.nan  # default value

    tau_g, sigma_g = adaptive_steps_goldstein(x, y, x_, y_, tau, sigma, memory, eta=0.95)
    alpha_g = memory[-1][2]
    memory.pop(-1)
    
    if counter > 4:
        res_x_far = memory[max(-len(memory),-counter-1)][3]
        res_y_far = memory[max(-len(memory),-counter-1)][4]
        tau_far = memory[max(-len(memory), -counter-1)][0]
        sigma_far = memory[max(-len(memory), -counter-1)][1]
        res_x_mid = memory[-counter//2][3]
        res_y_mid = memory[-counter//2][4]
        res_x_ref3 = memory[-3][3]
        res_y_ref3 = memory[-3][4]
        res_x_ref2 = memory[-2][3]
        res_y_ref2 = memory[-2][4]
        res_x_ref = memory[-1][3]
        res_y_ref = memory[-1][4]
        res = np.array(memory)[-counter//2:,3] / tau + np.array(memory)[-counter//2:,4] / sigma
        estimated_rate_1 = approximate_spectral_radius(res)
        estimated_rate_2 = np.sqrt(((res_x / tau + res_y/sigma)
                          / (res_x_mid/tau + res_y_mid/sigma))) ** (1./(counter//2))
        estimated_rate_3 =  np.sqrt((res_x / tau + res_y/sigma)
                                    / (res_x_ref/tau + res_y_ref/sigma))
        z = np.concatenate([x,y])
        R, Rb = recursion_matrix(tau, sigma)
        estimated_rate_4 = np.sqrt((res_x + res_y)/(res_x_ref+res_y_ref))  #np.max(np.abs(determine_largest_eigenvalues(R, R @ z - z, 1)))
        estimated_rate_5 = np.max(np.abs(determine_largest_eigenvalues(R, R @ z - z, 2)))
        # Aitken acceleration is not useful
        estimated_rate = estimated_rate_1
        if ( abs((1-estimated_rate) / (1-memory[-1][6]) - 1) < 1e-3
                and abs(estimated_rate - 2*memory[-1][6] + memory[-2][6])/(1-memory[-1][6])**2 < 1e-5
                and estimated_rate < 1
                and res_x/tau + res_y/sigma < delta * (res_x_far/tau_far + res_y_far/sigma_far) ):
            # stabilization of the rate and sufficient decrease
            if (len(memory) > counter + 1 and
                   memory[-counter][6] < estimated_rate):
                # the new rate is worse
                up_down = -up_down
            tau = tau * alpha**up_down
            sigma = sigma / alpha**up_down
            counter = 0
            goldstein_warm_up = False
    if goldstein_warm_up == True:   
        if abs(tau / tau_g - 1) > 1e-5:
            counter = 0
            tau = tau_g
            sigma = sigma_g
        #else:
        #    stop_warm_up = 1

    memory.append([tau, sigma, alpha_g, res_x, res_y, counter,
                   estimated_rate_1, estimated_rate_2, estimated_rate_3,
                   estimated_rate_4, estimated_rate_5, deriv_eigenvalue(tau)[0],
                   up_down])
    return tau, sigma

def adaptive_steps_residual_norm_goldstein_warm_up(x, y, x_, y_, tau, sigma, memory):
    return adaptive_steps_residual_norm(x, y, x_, y_, tau, sigma, memory,
                                        goldstein_warm_up=True)


def pdhg(x_init, y_init, tau0, sigma0, K, adaptive_steps=None,
         store_all_iterates=False):
    tau = tau0
    sigma = sigma0
    x = x_init.copy()
    y = y_init.copy()
    Nx = len(x)
    avgx = x.copy()
    avgy = y.copy()
    z_star = np.linalg.solve(M1, np.hstack([-c, b]))
    x_star = z_star[:Nx]
    y_star = z_star[Nx:]

    Nx = np.zeros(K)
    Ny = np.zeros(K)
    tV = np.zeros((K,4))
    memory = []
    if store_all_iterates == True:
        Z = np.zeros((len(z_star), K))
    
    for k in range(K):
        y_ = np.linalg.solve(Iy + sigma * S, y + sigma * (A @ x - b))
        x_ = np.linalg.solve(Ix + tau * Q, x - tau * (A.T @ (2*y_ - y) + c))

        if adaptive_steps is not None:
            tau, sigma = adaptive_steps(x, y, x_, y_,
                                        tau, sigma, memory)
        
        tV[k,0] = -tau/2. * norm(A.T @ (y_ - y))**2 + 0.5 * norm(x_ - x)**2/tau\
                + 0.5 * norm(y_ - y)**2/sigma
        tV[k,1] = L(x_, y_star) - L (x_star, y_)
        x_gamma = (x_star - tau * A.T @ y_) / (1+mup * tau)
        tV[k,2] = L(x_, y_star) + sigma / 2 * norm(A @ x_ - b)**2 \
                  + b.dot(y) - y @ A @ x_gamma - 0.5 * mup * norm(x_gamma)**2 \
                  - 0.5 / tau * norm(x_gamma - x_star)**2
        tV[k,3] = tau / 2 * norm(x_gamma - x_star)**2

        x = x_.copy()
        y = y_.copy()
        avgx = k/(k+1) * avgx + 1./(k+1)*x_
        avgy = k/(k+1) * avgy + 1./(k+1)*y_
    
        Nx[k] = norm(x - x_star)
        Ny[k] = norm(y - y_star)

        if store_all_iterates == True:
            Z[:,k] = np.concatenate([x,y])
        if Nx[k] + Ny[k] < 1e-14:
            break
    if store_all_iterates == True:
        return x, y, avgx, avgy, Nx, Ny, tV, memory, Z
    else:
        return x, y, avgx, avgy, Nx, Ny, tV, memory
        


def pdhg_V(x_init, y_init, tau, sigma, K):
    x = x_init.copy()
    y = y_init.copy()
    Nx = len(x)
    avgx = x.copy()
    avgy = y.copy()
    z_star = np.linalg.solve(M1, np.hstack([-c, b]))
    x_star = z_star[:Nx]
    y_star = z_star[Nx:]

    Nx = np.zeros(K)
    Ny = np.zeros(K)
    tV = np.zeros((K,4))
    
    for k in range(K):
        x_ = np.linalg.solve(Ix + tau * Q, x - tau * (A.T @ y + c))
        y_ = np.linalg.solve(Iy + sigma * S, y + sigma * (A @ x_ - b))
        tV[k,0] = -tau/2. * norm(A.T @ (y_ - y))**2 + 0.5 * norm(x_ - x)**2/tau\
                + 0.5 * norm(y_ - y)**2/sigma
        tV[k,1] = L(x_, y_star) - L (x_star, y_)
        x_gamma = (x_star - tau * A.T @ y_) / (1+mup * tau)
        tV[k,2] = L(x_, y_star) + sigma / 2 * norm(A @ x_ - b)**2 \
                  + b.dot(y) - y @ A @ x_gamma - 0.5 * mup * norm(x_gamma)**2 \
                  - 0.5 / tau * norm(x_gamma - x_star)**2
        tV[k,3] = tau / 2 * norm(x_gamma - x_star)**2
        
        x = x_ - tau * A.T @ (y_ - y)
        y = y_.copy()
        avgx = k/(k+1) * avgx + 1./(k+1)*x_
        avgy = k/(k+1) * avgy + 1./(k+1)*y_
    
        Nx[k] = norm(x - x_star)
        Ny[k] = norm(y - y_star)

        if Nx + Ny < 1e-14:
            break

    return x, y, avgx, avgy, Nx, Ny, tV


def rapdhg(x_init, y_init, tau, sigma, period, K):
    x = x_init.copy()
    y = y_init.copy()

    x_star = spl.spsolve(A, b)
    y_star = spl.spsolve(A.T, -c)
    
    Nx = np.zeros(K)
    Ny = np.zeros(K)
    ktot = 0

    while ktot < K:
        x_in, y_in, avgx, avgy, Nx_in, Ny_in, _ = pdhg(x, y,  tau, sigma,
                                                    min(period, K-ktot))
        x = avgx
        y = avgy
        # x = x - tau * period * (A.T @ avgy + c)  # stable mais lent
        # y = y + sigma * period * (A @ avgx - b)
        Nx[ktot:ktot+len(Nx_in)] = Nx_in
        Ny[ktot:ktot+len(Ny_in)] = Ny_in
        ktot = ktot + period

    return x, y, x_in, y_in, Nprimal, Ndual


#TAUS = 0.001 * 2 ** np.arange(20)
#z0 = np.arange(Nx+Ny)
#Goldstein = [primal_dual_residuals_ratio(tau, z0, K=1)[-1] for tau in TAUS]
#Deriv_eig = [deriv_eigenvalue(tau)[-1] for tau in TAUS]
#Deriv_res = [deriv_squared_residual(tau, z0) for tau in TAUS]

z0 = np.ones(Nx+Ny)
tau0 = 0.01 / norm(A)
sigma0 = gamma / tau0 / norm(A)**2
K = 3000
K1 = 3000



x_r, y_r, avgx_r, avgy_r, Nx_r, Ny_r, tV_r, memory_r = pdhg(z0[:Nx], z0[Nx:], tau0, sigma0, K, adaptive_steps_residual_norm)
memory_r = np.array(memory_r)

x_b, y_b, avgx_b, avgy_b, Nx_b, Ny_b, tV_b, memory_b = pdhg(z0[:Nx], z0[Nx:], tau0, sigma0, K1, adaptive_steps=None)

x, y, avgx, avgy, Nx_g, Ny_g, tV_g, memory_g = pdhg(z0[:Nx], z0[Nx:], tau0, sigma0, K1, adaptive_steps_goldstein)
memory_g = np.array(memory_g)

x_d, y_d, avgx_d, avgy_d, Nx_d, Ny_d, tV_d, memory_d = pdhg(z0[:Nx], z0[Nx:], tau0, sigma0, K1, adaptive_steps_derivative)
memory_d = np.array(memory_d)

x_rg, y_rg, avgx_rg, avgy_rg, Nx_rg, Ny_rg, tV_rg, memory_rg = pdhg(z0[:Nx], z0[Nx:], tau0, sigma0, K1, adaptive_steps_residual_norm_goldstein_warm_up)
memory_rg = np.array(memory_rg)

Versions = ['No adaptation',  #'Rate estimate (Alg. 5 $\\alpha_0=0$)', 'Residual balance and rate estimate (Alg. 5 $\\alpha_0>0$)',
            'Residual balance', # (Alg. 3)',
            'Direct spectral radius optimization']
plt.figure(1)
plt.semilogy(np.sqrt(Nx_b**2/tau0 + Ny_b**2/sigma0), '-')
#plt.semilogy(np.sqrt(Nx_r**2/tau0 + Ny_r**2/sigma0), '--')
#plt.semilogy(np.sqrt(Nx_rg**2/tau0 + Ny_rg**2/sigma0), '-')
plt.semilogy(np.sqrt(Nx_g**2/tau0 + Ny_g**2/sigma0), '-.', color='tab:red')
plt.semilogy(np.sqrt(Nx_d**2/tau0 + Ny_d**2/sigma0), ':', color='purple')
plt.ylabel('Distance to the saddle point $||z-z_*||_{V_0}$')
plt.xlabel('Iteration')
plt.legend(Versions)
plt.show()

plt.figure(4)
plt.semilogy(tau0 * np.ones(K), ':')
plt.semilogy(memory_r[:,0], '--')
plt.semilogy(memory_rg[:,0], '-')
plt.semilogy(memory_g[:,0], '-.')
plt.semilogy(memory_d[:,0], ':')
#plt.legend(Versions)
plt.ylabel('$τ_k$')
plt.xlabel('Iteration')
plt.show()

plt.figure(2)
plt.plot(np.abs(memory_rg[:,6]), '--')
plt.plot(np.abs(memory_rg[:,7]), '-.')
plt.plot(np.abs(memory_rg[:,8]), '-')
plt.plot(np.abs(memory_rg[:,11]), ':')
plt.ylabel('$|\lambda_1|$'); plt.xlabel('Iteration')
plt.legend(['cycles', 'long run', 'instant', 'true'])
plt.show()


plt.figure(3)
plt.plot(np.abs(memory_r[0:1000,8]), '-', color='tab:orange')
plt.plot(np.abs(memory_r[0:1000,9]), '-.', color='tab:green')
plt.plot(np.abs(memory_r[0:1000,11]), ':', color='tab:red')
plt.legend(['V-norm', '2-norm', 'true'])
plt.show()
